import 'dart:math' as math;

import 'package:knt_bloc/knt_bloc.dart';

import '../model/book.dart';
import '../repo/book_repo.dart';

class BookBloc extends BasePagingBloc<void, Book> with Frontend {
  final _books = <Book>[];
  bool _usingIsolator = false;

  BookBloc() {
    on<PrepareIsolateEvent>(
      _handlePrepareIsolateEvent,
      transformer: droppable(),
    );
    on<LoadBooksEvent>(handleLoadBooks);
  }

  @override
  bool get needLog => true;

  @override
  Future<BaseResponsePaging<Book>> pagingInProgress(
      BaseRequestPaging<void> pagingQuery) async {
    List<Book> partialBooks = _books.sublist(
      pagingQuery.offset,
      math.min(pagingQuery.offset + pagingQuery.limit, _books.length),
    );
    return BaseResponsePaging<Book>(dataList: partialBooks);
  }

  @override
  BasePagingResultState<Book> pagingStateInstance(
    List<Book> dataList,
    int offset, {
    String? tag,
  }) =>
      ShowBooksState(dataList, offset, tag: tag);

  Future<void> _handlePrepareIsolateEvent(
    PrepareIsolateEvent event,
    Emitter<BaseState> emit,
  ) async {
    await initBackend(initializer: _launch);
  }

  Future<void> handleLoadBooks(
    LoadBooksEvent event,
    Emitter<BaseState> emit,
  ) async {
    _usingIsolator = event.usingIsolate;
    emit(const BaseLoadingState());
    await Future.delayed(const Duration(milliseconds: 200));
    // throw const KntException(KntIssue.hostNotFound);
    if (_usingIsolator) {
      await initBackend(initializer: _launch);
      await run(event: event.tag);
      return;
    }
    final loadedBook = await BookRepo().fetchBooks();
    _setBooksFromBackend(event: (LoadBooksEvent).toString(), data: loadedBook);
  }

  static BookBlocBackend _launch(BackendArgument<void> argument) {
    return BookBlocBackend(argument);
  }

  void _setBooksFromBackend({required String event, required List<Book> data}) {
    _books.clear();
    _books.addAll(data);
    add(const BasePagingEvent<void>(isFirstRequest: true));
  }

  @override
  void initActions() {
    whenEventCome((LoadBooksEvent).toString()).run(_setBooksFromBackend);
  }
}

class BookBlocBackend extends Backend {
  final _repo = BookRepo();

  BookBlocBackend(BackendArgument<void> argument) : super(argument: argument);

  Future<void> _loadBooks({required String event, void data}) async {
    final books = await _repo.fetchBooks();
    await send(
      event: (LoadBooksEvent).toString(),
      data: books,
      sendDirectly: true,
    );
  }

  @override
  void initActions() {
    whenEventCome((LoadBooksEvent).toString()).run(_loadBooks);
  }
}

// region EVENT
class PrepareIsolateEvent extends BaseEvent {}

class LoadBooksEvent extends BaseEvent {
  final bool usingIsolate;

  const LoadBooksEvent({
    this.usingIsolate = false,
  });
}
// endregion

// region STATE
class ShowBooksState extends BasePagingResultState<Book> {
  const ShowBooksState(
    List<Book> dataList,
    int currentOffset, {
    String? tag,
  }) : super(dataList, currentOffset, tag: tag);
}
// endregion
